/* $Id: perfex.c,v 1.4 2001/03/31 15:17:38 mikpe Exp $
 * perfex.c
 *
 * NAME
 *	perfex - a command-line interface to x86 performance counters
 *
 * SYNOPSIS
 *	perfex [-e event] .. command
 *	perfex -i
 *
 * DESCRIPTION
 *	The given command is executed; after it is complete, perfex
 *	prints the values of the various hardware performance counters.
 *
 * OPTIONS
 *	-e event
 *		Specify an event to be counted.
 *		Multiple event specifiers may be given, limited by the
 *		number of available event selection registers (4 on K7,
 *		2 on P6, 1 on P5). Each event specifier is a 32-bit
 *		processor-dependent hexadecimal number. The order in
 *		which the events are specified is significant.
 *		The counts, together with an event description are written
 *		to stderr.
 *
 *	-i
 *		Instead of running a command, generate output which
 *		identifies the current processor and its capabilities.
 *
 * EXAMPLE
 *	The following command counts the number of retired instructions
 *	in user-mode on an Intel P6 processor:
 *
 *	perfex -e 0x004100C0 gcc
 *
 * DEPENDENCIES
 *	perfex only works on Linux/x86 systems which have been modified
 *	to include the perfctr driver. This driver is available at
 *	http://www.csd.uu.se/~mikpe/linux/perfctr/.
 *
 * BUGS
 *	The event description output should be the event specifier,
 *	but instead it is the pmc number. Fixing this requires changes
 *	to the kernel driver API.
 *
 * NOTES
 *	perfex is superficially similar to IRIX' perfex(1).
 *	The -a, -mp, -s, and -x options are not yet implemented.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */

/*
 * Theory of operation:
 * - Parent creates a socketpair().
 * - Parent forks.
 * - Child opens /proc/self/perfctr and sets up its perfctrs.
 * - Child sends its perfctr fd to parent via the socketpair().
 * - Child exec:s the command.
 * - Parent waits for child to exit.
 * - Parent receives child's perfctr fd via the socketpair().
 * - Parent mmap():s child's perfctr fd and reads the final counts.
 */

#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "libperfctr.h"
#define PAGE_SIZE	4096

/*
 * Our child-to-parent protocol is the following:
 * There is an int-sized data packet, with an optional 'struct cmsg_fd'
 * control message attached.
 * The data packet (which must be present, as control messages don't
 * work with zero-sized payloads) contains an 'int' status.
 * If status > 0, then it is the number of enabled counters.
 * If status < 0, then it is a negated 'errno' value from the child's
 * perfctr setup code.
 */

struct cmsg_fd {
    struct cmsghdr hdr;
    int fd;
};

int my_send(int sock, int fd, int status)
{
    struct msghdr msg;
    struct iovec iov;
    struct cmsg_fd cmsg_fd;
    int buf[1];

    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_flags = 0;

    /* man 7 unix says: "To pass file descriptors or credentials
       you need to send/read at least one byte." */
    buf[0] = status;
    iov.iov_base = buf;
    iov.iov_len = sizeof buf;
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    if( status < 0 ) {	/* -errno */
	msg.msg_control = 0;
	msg.msg_controllen = 0;
    } else {
	cmsg_fd.hdr.cmsg_len = sizeof cmsg_fd;
	cmsg_fd.hdr.cmsg_level = SOL_SOCKET;
	cmsg_fd.hdr.cmsg_type = SCM_RIGHTS;
	cmsg_fd.fd = fd;
	msg.msg_control = &cmsg_fd;
	msg.msg_controllen = sizeof cmsg_fd;
    }
    return sendmsg(sock, &msg, 0) == sizeof buf ? 0 : -1;
}

int my_send_err(int sock)
{
    return my_send(sock, -1, -errno);
}

int my_receive(int sock, int *fd, int *status)
{
    struct msghdr msg;
    struct iovec iov;
    struct cmsg_fd cmsg_fd;
    int buf[1];

    msg.msg_name = NULL;
    msg.msg_namelen = 0;
    msg.msg_flags = 0;

    buf[0] = -1;
    iov.iov_base = buf;
    iov.iov_len = sizeof(buf);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;

    memset(&cmsg_fd, ~0, sizeof cmsg_fd);
    msg.msg_control = &cmsg_fd;
    msg.msg_controllen = sizeof cmsg_fd;

    if( recvmsg(sock, &msg, 0) != sizeof buf )
	return -1;

    if( buf[0] >= 0 &&
	msg.msg_control == &cmsg_fd &&
	msg.msg_controllen == sizeof cmsg_fd &&
	cmsg_fd.hdr.cmsg_type == SCM_RIGHTS &&
	cmsg_fd.hdr.cmsg_level == SOL_SOCKET &&
	cmsg_fd.hdr.cmsg_len == sizeof cmsg_fd &&
	cmsg_fd.fd >= 0 ) {
	*fd = cmsg_fd.fd;
	*status = buf[0];
	return 0;
    }

    if( msg.msg_controllen == 0 && buf[0] < 0 )
	errno = -buf[0];
    else
	errno = EPROTO;
    return -1;
}

int do_child(int sock, const struct perfctr_control *control, char **argv)
{
    int fd;
    volatile const struct vperfctr_state *kstate;

    fd = open("/proc/self/perfctr", O_RDONLY);
    if( fd < 0 ) {
	my_send_err(sock);
	return 1;
    }
    kstate = mmap(NULL, PAGE_SIZE, PROT_READ, MAP_SHARED, fd, 0);
    if( kstate == MAP_FAILED ) {
	my_send_err(sock);
	return 1;
    }
    if( ioctl(fd, VPERFCTR_CONTROL, control) < 0 ) {
	my_send_err(sock);
	return 1;
    }
    /* we send kstate->status as a workaround for P5 and similar
       CPUs which don't have one evntsel per perfctr */
    if( my_send(sock, fd, kstate->status) < 0 ) {
	my_send_err(sock);	/* well, we can try.. */
	return 1;
    }
    munmap((void*)kstate, PAGE_SIZE);
    close(fd);
    close(sock);
    execvp(argv[0], argv);
    perror(argv[0]);
    return 1;
}

int do_parent(int sock, int child_pid)
{
    int child_status;
    int fd, status, i;
    volatile const struct vperfctr_state *kstate;

    /* this can be done before or after the recvmsg() */
    if( waitpid(child_pid, &child_status, 0) < 0 ) {
	perror("perfex: waitpid");
	return 1;
    }
    if( !WIFEXITED(child_status) ) {
	fprintf(stderr, "perfex: child did not exit normally\n");
	return 1;
    }
    if( my_receive(sock, &fd, &status) < 0 ) {
	perror("perfex: receiving fd/status");
	return 1;
    }
    close(sock);
    kstate = mmap(NULL, PAGE_SIZE, PROT_READ, MAP_SHARED, fd, 0);
    if( kstate == MAP_FAILED ) {
	perror("perfex: mmap");
	return 1;
    }
    close(fd);

    /* XXX: We need cpu_type here, in order to deal with P5 and
       P5-like processors. Fix this, and report evntsel instead
       of pmc number. */
    fprintf(stderr, "tsc\t0x%016llX\n", kstate->sum.ctr[0]);
    for(i = 1; i < status; ++i)
	fprintf(stderr, "pmc %u\t0x%016llX\n", i-1, kstate->sum.ctr[i]);

    munmap((void*)kstate, PAGE_SIZE);
    return WEXITSTATUS(child_status);
}

int do_perfex(const struct perfctr_control *control, char **argv)
{
    int pid;
    int sv[2];

    if( socketpair(AF_UNIX, SOCK_DGRAM, 0, sv) < 0 ) {
	perror("perfex: socketpair");
	return 1;
    }
    pid = fork();
    if( pid < 0 ) {
	perror("perfex: fork");
	return 1;
    }
    if( pid == 0 ) {
	close(sv[0]);
	return do_child(sv[1], control, argv);
    } else {
	close(sv[1]);
	return do_parent(sv[0], pid);
    }
}

int do_info(void)
{
    int fd;
    struct perfctr_info info;
    static const char * const features[] = { "rdpmc", "rdtsc", "pcint" };
    int fi;
    int comma;

    fd = open("/proc/self/perfctr", O_RDONLY);
    if( fd < 0 ) {
	perror("perfex: open /proc/self/perfctr");
	return 1;
    }
    if( ioctl(fd, PERFCTR_INFO, &info) != 0 ) {
	perror("perfex: PERFCTR_INFO");
	return 1;
    }
    close(fd);
    printf("perfex: driver\t\t%s\n", info.version);
    printf("perfex: nrcpus\t\t%u\n", info.nrcpus);
    printf("perfex: cpu type\t%u (%s)\n", info.cpu_type, perfctr_cpu_name(&info));
    printf("perfex: cpu speed\t%lu kHz\n", info.cpu_khz);
    printf("perfex: nrctrs\t\t%u\n", perfctr_cpu_nrctrs(&info));
    printf("perfex: features\t%#x (", info.cpu_features);
    for(comma = 0, fi = 0; fi < sizeof features / sizeof features[0]; ++fi) {
	unsigned fmask = 1 << fi;
	if( info.cpu_features & fmask ) {
	    if( comma )
		printf(",");
	    printf("%s", features[fi]);
	    comma = 1;
	}
    }
    printf(")\n");
    return 0;
}

int main(int argc, char **argv)
{
    struct perfctr_control control;
    unsigned long evntsel;
    int n;
    char *arg;
    char *endp;

    memset(&control, 0, sizeof control);
    n = 0;

    while( ++argv, --argc > 0 ) {
	arg = argv[0];
	if( arg[0] != '-' )
	    break;
	switch( arg[1] ) {
	  case 'i':
	    return do_info();
	  case 'e':
	    ++argv;
	    if( --argc <= 0 ) {
		fprintf(stderr, "perfex: event specifier missing\n");
		return 1;
	    }
	    if( n >= sizeof control.evntsel / sizeof control.evntsel[0] ) {
		fprintf(stderr, "perfex: too many event specifiers\n");
		return 1;
	    }
	    evntsel = strtoul(argv[0], &endp, 16);
	    if( endp[0] != '\0' ) {
		fprintf(stderr, "perfex: invalid number: '%s'\n", argv[0]);
		return 1;
	    }
	    control.evntsel[n] = evntsel;
	    ++n;
	    continue;
	}
	fprintf(stderr, "perfex: invalid option: '%s'\n", arg);
	return 1;
    }

    if( argc <= 0 ) {
	fprintf(stderr, "perfex: command missing\n");
	return 1;
    }

    return do_perfex(&control, argv);
}
