/* $Id: virtual.c,v 1.22 2001/04/07 21:52:17 mikpe Exp $
 * Virtual per-process performance counters.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/perfctr.h>

#include <asm/io.h>
#include <asm/uaccess.h>

#include "compat.h"
#include "virtual.h"

/****************************************************************
 *								*
 * Data types and macros.					*
 *								*
 ****************************************************************/

struct vperfctr {
/* User-visible fields: (must be first for mmap()) */
	struct vperfctr_state state;
/* Kernel-private fields: */
	atomic_t count;
	struct file *filp;	/* for filp sharing; currently doesn't work */
	unsigned iresume_imode;
};
#define IS_RUNNING(perfctr)	((perfctr)->state.status > 0)
#define IS_IMODE(perfctr)	((perfctr)->state.iresume.imode != 0)

/****************************************************************
 *								*
 * Resource management.						*
 *								*
 ****************************************************************/

/* XXX: perhaps relax this to number of _live_ perfctrs */
static spinlock_t nrctrs_lock = SPIN_LOCK_UNLOCKED;
int nrctrs = 0;
static const char this_service[] = __FILE__;
static void vperfctr_ihandler(unsigned long pc);

static int inc_nrctrs(void)
{
	const char *other;

	other = NULL;
	spin_lock(&nrctrs_lock);
	if( ++nrctrs == 1 )
		other = perfctr_cpu_reserve(this_service);
	spin_unlock(&nrctrs_lock);
	if( other ) {
		printk(KERN_ERR __FILE__
		       ": cannot operate, perfctr hardware taken by '%s'\n",
		       other);
		return -EBUSY;
	}
	perfctr_cpu_set_ihandler(vperfctr_ihandler);
	return 0;
}

static void dec_nrctrs(void)
{
	spin_lock(&nrctrs_lock);
	if( --nrctrs == 0 )
		perfctr_cpu_release(this_service);
	spin_unlock(&nrctrs_lock);
}

static struct vperfctr *vperfctr_alloc(void)
{
	unsigned long page;

	if( inc_nrctrs() != 0 )
		return NULL;
	page = get_zeroed_page(GFP_KERNEL);
	if( !page ) {
		dec_nrctrs();
		return NULL;
	}
	SetPageReserved(virt_to_page(page));
	return (struct vperfctr*) page;
}

static void vperfctr_free(struct vperfctr *perfctr)
{
	ClearPageReserved(virt_to_page(perfctr));
	free_page((unsigned long)perfctr);
	dec_nrctrs();
}

static struct vperfctr *get_empty_vperfctr(void)
{
	struct vperfctr *perfctr = vperfctr_alloc();
	if( perfctr ) {
		atomic_set(&perfctr->count, 1);
		perfctr->filp = NULL;
	}
	return perfctr;
}

static void put_vperfctr(struct vperfctr *perfctr)
{
	if( atomic_dec_and_test(&perfctr->count) )
		vperfctr_free(perfctr);
}

static unsigned int new_icookie(void)
{
	static spinlock_t lock = SPIN_LOCK_UNLOCKED;
	static unsigned int counter;
	int id;

	spin_lock(&lock);
	id = ++counter;
	spin_unlock(&lock);
	return id;
}

/****************************************************************
 *								*
 * Basic counter operations.					*
 *								*
 ****************************************************************/

/* Sample the CPU's registers to *now. Update accumulated counters. */
static __inline__ void
vperfctr_accumulate(struct vperfctr *perfctr, struct perfctr_low_ctrs *now)
{
	int i;

	perfctr_cpu_read_counters(perfctr->state.status, now);
	for(i = perfctr->state.status; --i >= 0;)
		perfctr->state.sum.ctr[i] +=
			now->ctr[i] - perfctr->state.start.ctr[i];
}

/* PRE: perfctr == TASK_VPERFCTR(current) && IS_RUNNING(perfctr)
 * Suspend the counters.
 */
static __inline__ void vperfctr_suspend(struct vperfctr *perfctr)
{
	if( IS_IMODE(perfctr) )
		perfctr_cpu_isuspend(&perfctr->state.iresume);
	/* XXX: shouldn't accumulate i-mode counters */
	vperfctr_accumulate(perfctr, &perfctr->state.iresume.ctrs);
	perfctr_cpu_disable_rdpmc();
}

/* PRE: perfctr == TASK_VPERFCTR(current) && IS_RUNNING(perfctr)
 * Restart the counters.
 */
static __inline__ void vperfctr_resume(struct vperfctr *perfctr)
{
	perfctr_cpu_enable_rdpmc();
	if( IS_IMODE(perfctr) )
		perfctr_cpu_iresume(&perfctr->state.iresume);
	perfctr_cpu_write_control(perfctr->state.status, &perfctr->state.control);
	/* XXX: shouldn't read i-mode counters */
	perfctr_cpu_read_counters(perfctr->state.status, &perfctr->state.start);
}

/* vperfctr interrupt handler (XXX: add buffering support) */
static void vperfctr_ihandler(unsigned long pc)
{
	struct task_struct *tsk = current;
	struct vperfctr *perfctr = TASK_VPERFCTR(tsk);
	siginfo_t si;

	if( !perfctr ) {
		printk(KERN_ERR __FUNCTION__ ": BUG! pid %d has no vperfctr\n", tsk->pid);
		return;
	}
	if( !IS_RUNNING(perfctr) ) {
		printk(KERN_ERR __FUNCTION__ ": BUG! pid %d's vperfctr has status %d, imode %#x\n",
		       tsk->pid, perfctr->state.status, perfctr->state.iresume.imode);
		return;
	}
	vperfctr_suspend(perfctr);
	perfctr->state.iresume_status = perfctr->state.status;
	perfctr->iresume_imode = perfctr->state.iresume.imode;
	perfctr->state.iresume.imode = 0;
	perfctr->state.status = 1;	/* only TSC remains active */
	perfctr->state.pc = pc;
	si.si_signo = perfctr->state.si_signo;
	si.si_errno = 0;
	si.si_code = perfctr->state.si_code;
	if( !send_sig_info(si.si_signo, &si, tsk) )
		send_sig(si.si_signo, tsk, 1);
}

/****************************************************************
 *								*
 * Process management callbacks.				*
 *								*
 ****************************************************************/

/* Called from from exit_thread() or sys_vperfctr_unlink().
 * Current has just detached its vperfctr.
 * If the counters are running, stop them and sample their final values.
 * Mark this perfctr as dead and decrement its use count.
 */
void __vperfctr_exit(struct vperfctr *perfctr)
{
	if( IS_RUNNING(perfctr) )
		vperfctr_suspend(perfctr);
	perfctr->state.status = VPERFCTR_STATUS_DEAD;
	put_vperfctr(perfctr);
}

/* schedule() --> switch_to() --> .. --> __vperfctr_suspend().
 * If the counters are running, sample their values.
 */
void __vperfctr_suspend(struct vperfctr *perfctr)
{
	if( IS_RUNNING(perfctr) )
		vperfctr_suspend(perfctr);
}

/* schedule() --> switch_to() --> .. --> __vperfctr_resume().
 * PRE: perfctr == TASK_VPERFCTR(current)
 * If the counters are runnable, resume them.
 */
void __vperfctr_resume(struct vperfctr *perfctr)
{
	if( IS_RUNNING(perfctr) )
		vperfctr_resume(perfctr);
}

/****************************************************************
 *								*
 * Virtual perfctr "system calls".				*
 *								*
 ****************************************************************/

/* PRE: perfctr == TASK_VPERFCTR(current) */
static int sys_vperfctr_stop(struct vperfctr *perfctr)
{
	if( IS_RUNNING(perfctr) ) {
		vperfctr_suspend(perfctr);
		perfctr->state.status = VPERFCTR_STATUS_STOPPED;
	}
	return 0;
}

static int
do_icontrol(struct vperfctr *perfctr, void *argp, int old_interface)
{
	struct vperfctr_icontrol arg;
	int status;
	int prev_status, i;
	unsigned int prev_start_tsc;
	unsigned int imode;

	if( old_interface ) {
		memset(&arg, 0, sizeof arg);
		if( copy_from_user(&arg.control, argp, sizeof arg.control) )
			return -EFAULT;
	} else {
		if( copy_from_user(&arg, argp, sizeof arg) )
			return -EFAULT;
	}
	/* XXX: i-mode and kernel-mode should be mutually exclusive */
	imode = 0;
	if( (status = perfctr_cpu_check_control(&arg.control, &imode)) < 0 )
		return status;
	if( old_interface && imode )
		return -EPERM;
	if( status == 0 )
		return sys_vperfctr_stop(perfctr);

	prev_status = perfctr->state.status;
	perfctr->state.status = status;
	perfctr->state.control = arg.control;
	perfctr->state.iresume.imode = imode;
	perfctr->state.iresume.ctrs = arg.ctrs;
	perfctr->state.iresume.icookie = new_icookie();
	perfctr->state.si_signo = arg.si_signo;
	perfctr->state.si_code = arg.si_code;

	/*
	 * Clear the perfctr sums and restart the perfctrs.
	 *
	 * If the counters were running before this control call,
	 * then don't clear the time-stamp counter's sum and don't
	 * overwrite its current start value.
	 */
	if( prev_status == 0 )
		perfctr->state.sum.ctr[0] = 0;
	for(i = 1; i < ARRAY_SIZE(perfctr->state.sum.ctr); ++i)
		perfctr->state.sum.ctr[i] = 0;
	prev_start_tsc = perfctr->state.start.ctr[0];
	vperfctr_resume(perfctr);	/* clobbers start.ctr[0] :-( */
	if( prev_status > 0 )
		perfctr->state.start.ctr[0] = prev_start_tsc;

	return 0;
}

/* Old interface */
static int
sys_vperfctr_control(struct vperfctr *perfctr, struct perfctr_control *argp)
{
	return do_icontrol(perfctr, (void*)argp, 1);
}

/* New interface */
static int
sys_vperfctr_icontrol(struct vperfctr *perfctr, struct vperfctr_icontrol *argp)
{
	return do_icontrol(perfctr, (void*)argp, 0);
}

static int sys_vperfctr_iresume(struct vperfctr *perfctr,
				struct perfctr_low_ctrs *argp)
{
	struct perfctr_low_ctrs ctrs;
	unsigned int prev_start_tsc;

	if( copy_from_user(&ctrs, argp, sizeof ctrs) )
		return -EFAULT;
	if( !IS_RUNNING(perfctr) )
		return -EINVAL;
	if( (perfctr->state.status = perfctr->state.iresume_status) <= 0 )
		return -EINVAL;
	perfctr->state.iresume_status = 0;
	perfctr->state.iresume.imode = perfctr->iresume_imode;
	perfctr->state.iresume.ctrs = ctrs;
	perfctr->state.iresume.icookie = new_icookie();
	prev_start_tsc = perfctr->state.start.ctr[0];
	vperfctr_resume(perfctr);
	perfctr->state.start.ctr[0] = prev_start_tsc;
	return 0;
}

/* PRE: perfctr == TASK_VPERFCTR(current) */
static int sys_vperfctr_unlink(struct vperfctr *perfctr)
{
	TASK_VPERFCTR(current) = NULL;
	__vperfctr_exit(perfctr);
	return 0;
}

/* PRE: perfctr == TASK_VPERFCTR(current)
 * Sample the current process' counters and update state.
 * This operation is used on processors like the pre-MMX Intel P5,
 * which cannot sample the counter registers in user-mode.
 */
static int sys_vperfctr_sample(struct vperfctr *perfctr)
{
	if( IS_RUNNING(perfctr) ) {
		struct perfctr_low_ctrs now;
		vperfctr_accumulate(perfctr, &now);
		perfctr->state.start = now;
	}
	return 0;
}

/****************************************************************
 *								*
 * Virtual perfctr file operations.				*
 *								*
 ****************************************************************/

static int vperfctr_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct vperfctr *perfctr;

	/* Only allow read-only mapping of first page. */
	if( (vma->vm_end - vma->vm_start) != PAGE_SIZE ||
	    vma_pgoff(vma) != 0 ||
	    (pgprot_val(vma->vm_page_prot) & _PAGE_RW) ||
	    (vma->vm_flags & (VM_WRITE | VM_MAYWRITE)) )
		return -EPERM;
	perfctr = filp->private_data;
	return remap_page_range(vma->vm_start, virt_to_phys(perfctr),
				PAGE_SIZE, vma->vm_page_prot);
}

static int vperfctr_release(struct inode *inode, struct file *filp)
{
	struct vperfctr *perfctr = filp->private_data;
	/*
	 * With the addition of /proc/self/perfctr, there is no longer
	 * a one-to-one perfctr <--> filp correspondence. This may
	 * change again when /proc/self/perfctr is reimplemented as
	 * a symbolic link to "perfctr:[...]".
	 */
	if( filp == perfctr->filp )
		perfctr->filp = NULL;
	filp->private_data = NULL;
	put_vperfctr(perfctr);
	return 0;
}

static int vperfctr_ioctl(struct inode *inode, struct file *filp,
			  unsigned int cmd, unsigned long arg)
{
	struct vperfctr *perfctr = filp->private_data;

	if( perfctr != TASK_VPERFCTR(current) )
		return -EPERM;
	switch( cmd ) {
	case PERFCTR_INFO:
		return sys_perfctr_info((struct perfctr_info*)arg);
	case VPERFCTR_CONTROL:
		return sys_vperfctr_control(perfctr, (struct perfctr_control*)arg);
	case VPERFCTR_STOP:
		return sys_vperfctr_stop(perfctr);
	case VPERFCTR_UNLINK:
		return sys_vperfctr_unlink(perfctr);
	case VPERFCTR_SAMPLE:
		return sys_vperfctr_sample(perfctr);
	case VPERFCTR_ICONTROL:
		return sys_vperfctr_icontrol(perfctr, (struct vperfctr_icontrol*)arg);
	case VPERFCTR_IRESUME:
		return sys_vperfctr_iresume(perfctr, (struct perfctr_low_ctrs*)arg);
	}
	return -EINVAL;
}

static int vperfctr_open(struct inode *inode, struct file *filp)
{
	struct task_struct *tsk;
	struct vperfctr *perfctr;

	tsk = current;
	if( !proc_pid_inode_denotes_task(inode, tsk) )
		return -EPERM;
	perfctr = TASK_VPERFCTR(tsk);
	if( !perfctr ) {
		perfctr = get_empty_vperfctr();
		if( !perfctr )
			return -ENOMEM;
		perfctr->filp = filp;
	}
	filp->private_data = perfctr;
	atomic_inc(&perfctr->count);
	if( !TASK_VPERFCTR(tsk) )
		TASK_VPERFCTR(tsk) = perfctr;
	return 0;
}

static struct file_operations vperfctr_file_ops = {
	OWNER_THIS_MODULE
	.mmap = vperfctr_mmap,
	.release = vperfctr_release,
	.ioctl = vperfctr_ioctl,
	.open = vperfctr_open,
};

/* XXX: inode->u.proc_i.op.proc_get_link() ought to take a fourth 'follow'
   parameter indicating whether readlink() or follow_link() is the caller */

#if !defined(MODULE)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)

void perfctr_set_proc_pid_ops(struct inode *inode)
{
	inode->i_fop = &vperfctr_file_ops;
}

#else	/* 2.2 */

#include <linux/proc_fs.h>

struct inode_operations perfctr_proc_pid_inode_operations = {
	.default_file_ops = &vperfctr_file_ops,
	.permission = proc_permission,
};

#endif	/* LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) */

#endif	/* !defined(MODULE) */

/****************************************************************
 *								*
 * module_init/exit						*
 *								*
 ****************************************************************/

#ifdef MODULE
static struct vperfctr_stub off;

static void vperfctr_stub_init(void)
{
	write_lock(&vperfctr_stub_lock);
	off = vperfctr_stub;
	vperfctr_stub.exit = __vperfctr_exit;
	vperfctr_stub.suspend = __vperfctr_suspend;
	vperfctr_stub.resume = __vperfctr_resume;
	vperfctr_stub.file_ops = &vperfctr_file_ops;
	write_unlock(&vperfctr_stub_lock);
}

static void vperfctr_stub_exit(void)
{
	write_lock(&vperfctr_stub_lock);
	vperfctr_stub = off;
	write_unlock(&vperfctr_stub_lock);
}
#else
static __inline__ void vperfctr_stub_init(void) { }
static __inline__ void vperfctr_stub_exit(void) { }
#endif	/* MODULE */

int __init vperfctr_init(void)
{
	vperfctr_stub_init();
	return 0;
}

void __exit vperfctr_exit(void)
{
	vperfctr_stub_exit();
}
