/* $Id: x86.c,v 1.17 2001/04/04 17:49:51 mikpe Exp $
 * x86 performance-monitoring counters driver.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/perfctr.h>

#include <asm/msr.h>

#include "compat.h"
#include "x86_compat.h"
#include "x86_tests.h"

/* Support for lazy evntsel MSR updates. */
struct perfctr_cpu_state {
	struct perfctr_control control;
	unsigned long icookie;	/* for optimising iresume */
} __attribute__((__aligned__(SMP_CACHE_BYTES)));
static struct perfctr_cpu_state per_cpu_state[NR_CPUS] __cacheline_aligned;

/* Intel P5, Cyrix 6x86MX/MII/III, Centaur WinChip C6/2/3 */
#define MSR_P5_CESR		0x11
#define MSR_P5_CTR0		0x12
#define MSR_P5_CTR1		0x13

/* Intel P6, VIA Cyrix III / C3 */
#define MSR_P6_PERFCTR0		0xC1
#define MSR_P6_PERFCTR1		0xC2
#define MSR_P6_EVNTSEL0		0x186
#define MSR_P6_EVNTSEL1		0x187

/* AMD K7 */
#define MSR_K7_EVNTSEL0		0xC0010000	/* .. 0xC0010003 */
#define MSR_K7_PERFCTR0		0xC0010004	/* .. 0xC0010007 */

#define rdmsrl(msr,low) \
	__asm__ __volatile__("rdmsr" : "=a"(low) : "c"(msr) : "edx")
#define rdpmcl(ctr,low) \
	__asm__ __volatile__("rdpmc" : "=a"(low) : "c"(ctr) : "edx")

/****************************************************************
 *								*
 * Driver procedures.						*
 *								*
 ****************************************************************/

static int p5_check_control(const struct perfctr_control *control,
			    unsigned int *imode)
{
	/* protect reserved and pin control bits */
	if( control->evntsel[0] & 0xFE00FE00 )
		return -EPERM;
	/* CTR1 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x00C00000 )
		return 3;
	/* CTR0 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x000000C0 )
		return 2;
	/* Only TSC is on. */
	return 1;
}

static void p5_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	if( nrctrs <= 1 )	/* no evntsel to write if only TSC is on */
		return;
	cpu = &per_cpu_state[smp_processor_id()].control;
	evntsel = control->evntsel[0];
	if( cpu->evntsel[0] != evntsel ) {
		cpu->evntsel[0] = evntsel;
		wrmsr(MSR_P5_CESR, evntsel, 0);
	}
}

static void p5_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->ctr[0]);
	if( nrctrs >= 2 ) {
		rdmsrl(MSR_P5_CTR0, ctrs->ctr[1]);
		if( nrctrs == 3 )
			rdmsrl(MSR_P5_CTR1, ctrs->ctr[2]);
	}
}

static void p5mmx_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->ctr[0]);
	if( nrctrs >= 2 ) {
		rdpmcl(0, ctrs->ctr[1]);
		if( nrctrs == 3 )
			rdpmcl(1, ctrs->ctr[2]);
	}
}

static void p5_clear_counters(void)
{
	wrmsr(MSR_P5_CESR, 0, 0);
	wrmsr(MSR_P5_CTR0, 0, 0);
	wrmsr(MSR_P5_CTR1, 0, 0);
}

static int p6_check_control(const struct perfctr_control *control,
			    unsigned int *imode)
{
	unsigned imode_mask, i;
	int last_pmc_on;

	last_pmc_on = -1;
	imode_mask = 0;
	for(i = 0; i < 2; ++i) {
		/* protect reserved and pin control bits */
		if( control->evntsel[i] & 0x00280000 )
			return -EPERM;
		/* only evntsel[0] has the ENable bit */
		if( i > 0 && (control->evntsel[i] & 0x00400000) )
			return -EPERM;
		/* check CPL field */
		if( !(control->evntsel[i] & 0x00030000) ) {
			/* not enabled: disallow random junk */
			if( control->evntsel[i] != 0 )
				return -EPERM;
			continue;
		}
		last_pmc_on = i;
		/* check INT bit */
		if( control->evntsel[i] & 0x00100000 )
			imode_mask |= (1 << i);
	}
	/* disallow i-mode counters if we cannot catch the interrupts */
	if( imode_mask && !(perfctr_info.cpu_features & PERFCTR_FEATURE_PCINT) )
		return -EPERM;
	/* check global enable bit */
	if( !(control->evntsel[0] & 0x00400000) ) {
		imode_mask = 0;
		last_pmc_on = -1;
	}
	/* adjust for TSC being ctr[0] */
	*imode = imode_mask << 1;
	return last_pmc_on + 2;
}

static void p6_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	cpu = &per_cpu_state[smp_processor_id()].control;
	if( nrctrs == 3 ) {
		evntsel = control->evntsel[1];
		if( evntsel != cpu->evntsel[1] ) {
			cpu->evntsel[1] = evntsel;
			wrmsr(MSR_P6_EVNTSEL1, evntsel, 0);
		}
	}	
	if( nrctrs >= 2 ) {
		evntsel = control->evntsel[0];
		if( evntsel != cpu->evntsel[0] ) {
			cpu->evntsel[0] = evntsel;
			wrmsr(MSR_P6_EVNTSEL0, evntsel, 0);
		}
	}
}

static void p6_clear_counters(void)
{
	int i;

	for(i = 0; i < 2; ++i) {
		wrmsr(MSR_P6_EVNTSEL0+i, 0, 0);
		wrmsr(MSR_P6_PERFCTR0+i, 0, 0);
	}
}

#if PERFCTR_INTERRUPT_SUPPORT
static void p6_isuspend(const struct perfctr_iresume *iresume)
{
	struct perfctr_cpu_state *state;

	/* XXX: should we or our caller check this? */
	if( !iresume->imode )
		return;
	state = &per_cpu_state[smp_processor_id()];
	state->control.evntsel[0] = 0;
	wrmsr(MSR_P6_EVNTSEL0, 0, 0);
	state->icookie = iresume->icookie;
}

static void p6_iresume(const struct perfctr_iresume *iresume)
{
	struct perfctr_cpu_state *state;
	unsigned imode;
	unsigned i;

	imode = iresume->imode;
	/* XXX: should we or our caller check this? */
	if( !imode )
		return;
	state = &per_cpu_state[smp_processor_id()];
	if( state->control.evntsel[0] ) {
		state->control.evntsel[0] = 0;
		wrmsr(MSR_P6_EVNTSEL0, 0, 0);
		state->icookie = -1;
	} else if( state->icookie == iresume->icookie )
		return;
	for(i = 1; (imode >>= 1) != 0; ++i)
		if( imode & 1 )
			wrmsr(MSR_P6_PERFCTR0-1+i, iresume->ctrs.ctr[i], -1);
}
#endif	/* PERFCTR_INTERRUPT_SUPPORT */

static int k7_check_control(const struct perfctr_control *control,
			    unsigned int *imode)
{
	unsigned imode_mask, i;
	int last_pmc_on;

	last_pmc_on = -1;
	imode_mask = 0;
	for(i = 0; i < 4; ++i) {
		/* protect reserved and pin control bits */
		if( control->evntsel[i] & 0x00280000 )
			return -EPERM;
		/* check enable bit and CPL field */
		if( !(control->evntsel[i] & 0x00400000) ||
		    !(control->evntsel[i] & 0x00030000) ) {
			/* not enabled: disallow random junk */
			if( control->evntsel[i] != 0 )
				return -EPERM;
			continue;
		}
		last_pmc_on = i;
		/* check INT bit */
		if( control->evntsel[i] & 0x00100000 )
			imode_mask |= (1 << i);
	}
	/* disallow i-mode counters if we cannot catch the interrupts */
	if( imode_mask && !(perfctr_info.cpu_features & PERFCTR_FEATURE_PCINT) )
		return -EPERM;
	/* adjust for TSC being ctr[0] */
	*imode = imode_mask << 1;
	return last_pmc_on + 2;
}

static void k7_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	int i;
	unsigned evntsel;

	cpu = &per_cpu_state[smp_processor_id()].control;
	for(i = nrctrs - 1; --i >= 0;) {	/* -1 for TSC */
		evntsel = control->evntsel[i];
		if( evntsel != cpu->evntsel[i] ) {
			cpu->evntsel[i] = evntsel;
			wrmsr(MSR_K7_EVNTSEL0+i, evntsel, 0);
		}
	}
}

static void k7_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	int i;

	rdtscl(ctrs->ctr[0]);
	for(i = nrctrs - 1; --i >= 0;)	/* -1 for TSC */
		rdpmcl(i, ctrs->ctr[i+1]);
}

static void k7_clear_counters(void)
{
	int i;

	for(i = 0; i < 4; ++i) {
		wrmsr(MSR_K7_EVNTSEL0+i, 0, 0);
		wrmsr(MSR_K7_PERFCTR0+i, 0, 0);
	}
}

#if PERFCTR_INTERRUPT_SUPPORT
static void k7_isuspend(const struct perfctr_iresume *iresume)
{
	struct perfctr_cpu_state *state;
	unsigned imode;
	unsigned i;

	imode = iresume->imode;
	/* XXX: should we or our caller check this? */
	if( !imode )
		return;
	state = &per_cpu_state[smp_processor_id()];
	for(i = 1; (imode >>= 1) != 0; ++i) {
		if( imode & 1 ) {
			state->control.evntsel[i-1] = 0;
			wrmsr(MSR_K7_EVNTSEL0-1+i, 0, 0);
		}
	}
	state->icookie = iresume->icookie;
}

static void k7_iresume(const struct perfctr_iresume *iresume)
{
	struct perfctr_cpu_state *state;
	unsigned imode;
	unsigned i;
	int icookie_valid;

	imode = iresume->imode;
	/* XXX: should we or our caller check this? */
	if( !imode )
		return;
	state = &per_cpu_state[smp_processor_id()];

	icookie_valid = 1;
	for(i = 1; (imode >>= 1) != 0; ++i) {
		if( (imode & 1) && state->control.evntsel[i-1] ) {
			state->control.evntsel[i-1] = 0;
			wrmsr(MSR_K7_EVNTSEL0-1+i, 0, 0);
			icookie_valid = 0;
		}
	}

	if( icookie_valid && state->icookie == iresume->icookie )
		return;

	imode = iresume->imode;
	for(i = 1; (imode >>= 1) != 0; ++i)
		if( imode & 1 )
			wrmsr(MSR_K7_PERFCTR0-1+i, iresume->ctrs.ctr[i], -1);
}
#endif	/* PERFCTR_INTERRUPT_SUPPORT */

static int mii_check_control(const struct perfctr_control *control,
			     unsigned int *imode)
{
	/* Protect reserved and pin control bits.
	 * CESR bits 9 and 25 are reserved in the Cyrix III,
	 * but not in the earlier processors where they control
	 * the external PM pins. In either case, we do not
	 * allow the user to touch them.
	 */
	if( control->evntsel[0] & 0xFA00FA00 )
		return -EPERM;
	/* CTR1 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x00C00000 )
		return 3;
	/* CTR0 is on if its CPL field is non-zero */
	if( control->evntsel[0] & 0x000000C0 )
		return 2;
	/* Only TSC is on. */
	return 1;
}

#if defined(CONFIG_PERFCTR_WINCHIP)
static int c6_check_control(const struct perfctr_control *control,
			    unsigned int *imode)
{
	/* protect reserved bits */
	if( control->evntsel[0] & 0xFF00FF00 )
		return -EPERM;
	return 3;	/* fake TSC and two perfctrs */
}

static void c6_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	ctrs->ctr[0] = 0;
	rdpmcl(0, ctrs->ctr[1]);
	rdpmcl(1, ctrs->ctr[2]);
}
#endif	/* CONFIG_PERFCTR_WINCHIP */

static int vciii_check_control(const struct perfctr_control *control,
			       unsigned int *imode)
{
	/* Protect reserved bits */
	if( control->evntsel[0] & ~0x1FF )
		return -EPERM;
	/* CTR1 is on if the event selector is non-zero */
	if( control->evntsel[0] )
		return 2;
	/* Only TSC is on */
	return 1;
}

static void vciii_write_control(int nrctrs, const struct perfctr_control *control)
{
	struct perfctr_control *cpu;
	unsigned evntsel;

	if( nrctrs < 2 )
		return;
	cpu = &per_cpu_state[smp_processor_id()].control;
	evntsel = control->evntsel[0];
	if( cpu->evntsel[0] != evntsel ) {
		unsigned int old, dummy;
		/* XXX: read-modify-write to preserve reserved bits;
		   hopefully this can go away after testing on real HW */
		cpu->evntsel[0] = evntsel;
		rdmsr(MSR_P6_EVNTSEL1, old, dummy);
		evntsel |= (old & ~0x1FF);
		wrmsr(MSR_P6_EVNTSEL1, evntsel, 0);
	}
}

static void vciii_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->ctr[0]);
	if( nrctrs == 2 )
		rdpmcl(1, ctrs->ctr[1]);
}

static int generic_check_control(const struct perfctr_control *control,
				 unsigned int *imode)
{
	return 1;
}

static void generic_write_control(int nrctrs, const struct perfctr_control *control)
{
}

static void generic_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->ctr[0]);
}

static void generic_clear_counters(void)
{
}

static void redirect_call(void *ra, void *to)
{
	/* we can only redirect `call near relative' instructions */
	if( *((unsigned char*)ra - 5) != 0xE8 ) {
		printk(KERN_WARNING __FILE__ ":" __FUNCTION__
		       ": unable to redirect caller %p to %p\n",
		       ra, to);
		return;
	}
	*(int*)((char*)ra - 4) = (char*)to - (char*)ra;
}

static int (*check_control)(const struct perfctr_control*, unsigned int*);
int perfctr_cpu_check_control(const struct perfctr_control *control,
			      unsigned int *imode)
{
	redirect_call(__builtin_return_address(0), check_control);
	return check_control(control, imode);
}

static void (*write_control)(int, const struct perfctr_control*);
void perfctr_cpu_write_control(int nrctrs, const struct perfctr_control *control)
{
	redirect_call(__builtin_return_address(0), write_control);
	return write_control(nrctrs, control);
}

static void (*read_counters)(int, struct perfctr_low_ctrs*);
void perfctr_cpu_read_counters(int nrctrs, struct perfctr_low_ctrs *ctrs)
{
	redirect_call(__builtin_return_address(0), read_counters);
	return read_counters(nrctrs, ctrs);
}

static void (*clear_counters)(void);
static void perfctr_cpu_clear_counters(void)
{
	return clear_counters();
}

#if PERFCTR_INTERRUPT_SUPPORT
static void (*cpu_isuspend)(const struct perfctr_iresume*);
void perfctr_cpu_isuspend(const struct perfctr_iresume *iresume)
{
	redirect_call(__builtin_return_address(0), cpu_isuspend);
	return cpu_isuspend(iresume);
}

static void (*cpu_iresume)(const struct perfctr_iresume*);
void perfctr_cpu_iresume(const struct perfctr_iresume *iresume)
{
	redirect_call(__builtin_return_address(0), cpu_iresume);
	return cpu_iresume(iresume);
}
#endif	/* PERFCTR_INTERRUPT_SUPPORT */

/****************************************************************
 *								*
 * Processor detection and initialisation procedures.		*
 *								*
 ****************************************************************/

static int __init intel_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 5:
		if( cpu_has_mmx ) {
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_P5MMX;
			read_counters = p5mmx_read_counters;
		} else {
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_P5;
			perfctr_info.cpu_features &= ~PERFCTR_FEATURE_RDPMC;
			read_counters = p5_read_counters;
		}
		write_control = p5_write_control;
		check_control = p5_check_control;
		clear_counters = p5_clear_counters;
		perfctr_p5_init_tests();
		return 0;
	case 6:
		if( boot_cpu_data.x86_model >= 7 )	/* PIII */
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_PIII;
		else if( boot_cpu_data.x86_model >= 3 )	/* PII or Celeron */
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_PII;
		else
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_P6;
		read_counters = p5mmx_read_counters;
		write_control = p6_write_control;
		check_control = p6_check_control;
		clear_counters = p6_clear_counters;
#if PERFCTR_INTERRUPT_SUPPORT
		if( cpu_has_apic ) {
			perfctr_info.cpu_features |= PERFCTR_FEATURE_PCINT;
			cpu_isuspend = p6_isuspend;
			cpu_iresume = p6_iresume;
		}
#endif
		perfctr_p6_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init amd_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* K7. Model 1 does not have a local APIC. */
		perfctr_info.cpu_type = PERFCTR_X86_AMD_K7;
		read_counters = k7_read_counters;
		write_control = k7_write_control;
		check_control = k7_check_control;
		clear_counters = k7_clear_counters;
#if PERFCTR_INTERRUPT_SUPPORT
		if( cpu_has_apic ) {
			perfctr_info.cpu_features |= PERFCTR_FEATURE_PCINT;
			cpu_isuspend = k7_isuspend;
			cpu_iresume = k7_iresume;
		}
#endif
		perfctr_k7_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init cyrix_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* 6x86MX, MII, or III */
		perfctr_info.cpu_type = PERFCTR_X86_CYRIX_MII;
		read_counters = p5mmx_read_counters;
		write_control = p5_write_control;
		check_control = mii_check_control;
		clear_counters = p5_clear_counters;
		perfctr_mii_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init centaur_init(void)
{
	switch( boot_cpu_data.x86 ) {
#if defined(CONFIG_PERFCTR_WINCHIP)
	case 5:
		switch( boot_cpu_data.x86_model ) {
		case 4: /* WinChip C6 */
			perfctr_info.cpu_type = PERFCTR_X86_WINCHIP_C6;
			break;
		case 8: /* WinChip 2, 2A, or 2B */
		case 9: /* WinChip 3, a 2A with larger cache and lower voltage */
			perfctr_info.cpu_type = PERFCTR_X86_WINCHIP_2;
			break;
		default:
			return -ENODEV;
		}
		/*
		 * TSC must be inaccessible for perfctrs to work.
		 */
		if( !(read_cr4() & X86_CR4_TSD) || cpu_has_tsc )
			return -ENODEV;
		perfctr_info.cpu_features &= ~PERFCTR_FEATURE_RDTSC;
		read_counters = c6_read_counters;
		write_control = p5_write_control;
		check_control = c6_check_control;
		clear_counters = p5_clear_counters;
		perfctr_c6_init_tests();
		return 0;
#endif
	case 6: /* VIA Cyrix III / C3 */
		if( !cpu_has_tsc )
			return -ENODEV;
		switch( boot_cpu_data.x86_model ) {
		case 6:	/* VIA Cyrix III */
		case 7:	/* VIA C3 */
			break;
		default:
			return -ENODEV;
		}
		perfctr_info.cpu_type = PERFCTR_X86_VIA_CYRIX_III;
		read_counters = vciii_read_counters;
		write_control = vciii_write_control;
		check_control = vciii_check_control;
		clear_counters = generic_clear_counters;
		perfctr_vciii_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init generic_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	perfctr_info.cpu_features &= ~PERFCTR_FEATURE_RDPMC;
	perfctr_info.cpu_type = PERFCTR_X86_GENERIC;
	check_control = generic_check_control;
	write_control = generic_write_control;
	read_counters = generic_read_counters;
	clear_counters = generic_clear_counters;
	return 0;
}

static char generic_name[] __initdata = "Generic x86 with TSC";
static char p5_name[] __initdata = "Intel Pentium";
static char p5mmx_name[] __initdata = "Intel Pentium MMX";
static char p6_name[] __initdata = "Intel Pentium Pro";
static char pii_name[] __initdata = "Intel Pentium II";
static char piii_name[] __initdata = "Intel Pentium III";
static char mii_name[] __initdata = "Cyrix 6x86MX/MII/III";
static char wcc6_name[] __initdata = "WinChip C6";
static char wc2_name[] __initdata = "WinChip 2/3";
static char k7_name[] __initdata = "AMD K7";
static char vciii_name[] __initdata = "VIA Cyrix III / C3";

char *perfctr_cpu_name[] __initdata = {
	[PERFCTR_X86_GENERIC] generic_name,
	[PERFCTR_X86_INTEL_P5] p5_name,
	[PERFCTR_X86_INTEL_P5MMX] p5mmx_name,
	[PERFCTR_X86_INTEL_P6] p6_name,
	[PERFCTR_X86_INTEL_PII] pii_name,
	[PERFCTR_X86_INTEL_PIII] piii_name,
	[PERFCTR_X86_CYRIX_MII] mii_name,
	[PERFCTR_X86_WINCHIP_C6] wcc6_name,
	[PERFCTR_X86_WINCHIP_2] wc2_name,
	[PERFCTR_X86_AMD_K7] k7_name,
	[PERFCTR_X86_VIA_CYRIX_III] vciii_name,
};

static void __init perfctr_cpu_init_one(void *ignore)
{
	perfctr_cpu_clear_counters();
#if PERFCTR_INTERRUPT_SUPPORT
	if( cpu_has_apic )
		apic_write(APIC_LVTPC, LOCAL_PERFCTR_VECTOR);
#endif
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		set_in_cr4_local(X86_CR4_PCE);
}

static void __exit perfctr_cpu_exit_one(void *ignore)
{
	perfctr_cpu_clear_counters();
#if PERFCTR_INTERRUPT_SUPPORT
	if( cpu_has_apic )
		apic_write(APIC_LVTPC, APIC_DM_NMI | APIC_LVT_MASKED);
#endif
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		clear_in_cr4_local(X86_CR4_PCE);
}

#if defined(NMI_LOCAL_APIC) && defined(CONFIG_PM)

static void __init unregister_nmi_pmdev(void)
{
	if( nmi_pmdev ) {
		apic_pm_unregister(nmi_pmdev);
		nmi_pmdev = 0;
	}
}

static int x86_pm_callback(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	/* XXX: incomplete */
	return 0;
}

static struct pm_dev *x86_pmdev;

static void __init x86_pm_init(void)
{
	x86_pmdev = apic_pm_register(PM_SYS_DEV, 0, x86_pm_callback);
}

static void __exit x86_pm_exit(void)
{
	if( x86_pmdev ) {
		apic_pm_unregister(x86_pmdev);
		x86_pmdev = NULL;
	}
}

#else

static __inline__ void unregister_nmi_pmdev(void) { }
static __inline__ void x86_pm_init(void) { }
static __inline__ void x86_pm_exit(void) { }

#endif	/* NMI_LOCAL_APIC && CONFIG_PM */

#if defined(NMI_LOCAL_APIC)

static void __init disable_nmi_watchdog(void)
{
	if( nmi_perfctr_msr ) {
		nmi_perfctr_msr = 0;
		printk(KERN_NOTICE "perfctr: disabled nmi_watchdog\n");
		unregister_nmi_pmdev();
	}
}

#else

static __inline__ void disable_nmi_watchdog(void) { }

#endif

int __init perfctr_cpu_init(void)
{
	int err = -ENODEV;

	/* RDPMC and RDTSC are on by default. They will be disabled
	   by the init procedures if necessary. */
	perfctr_info.cpu_features = PERFCTR_FEATURE_RDPMC | PERFCTR_FEATURE_RDTSC;

	if( cpu_has_msr ) {
		switch( boot_cpu_data.x86_vendor ) {
		case X86_VENDOR_INTEL:
			err = intel_init();
			break;
		case X86_VENDOR_AMD:
			err = amd_init();
			break;
		case X86_VENDOR_CYRIX:
			err = cyrix_init();
			break;
		case X86_VENDOR_CENTAUR:
			err = centaur_init();
		}
	}
	if( err ) {
		err = generic_init();	/* last resort */
		if( err )
			return err;
	}
	/*
	 * Put the hardware in a sane state:
	 * - clear perfctr MSRs
	 * - set up APIC_LVTPC
	 * - set CR4.PCE [on permanently due to __flush_tlb_global()]
	 * - install our default interrupt handler
	 */
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		mmu_cr4_features |= X86_CR4_PCE;
	perfctr_cpu_init_one(NULL);
	smp_call_function(perfctr_cpu_init_one, NULL, 1, 1);
	perfctr_cpu_set_ihandler(NULL);
	/*
	 * Fix up the connection to the local APIC:
	 * - disable and disconnect the NMI watchdog
	 * - register our PM callback
	 */
	disable_nmi_watchdog();
	x86_pm_init();
	/*
	 * per_cpu_state[] is initialised to contain "impossible"
	 * evntsel values guaranteed to differ from anything accepted
	 * by perfctr_cpu_check_control(). This way, initialisation of
	 * a CPU's evntsel MSRs will happen automatically the first time
	 * perfctr_cpu_write_control() executes on it.
	 * All-bits-one works for all currently supported processors.
	 * The memset also sets the icookies to -1, which is intentional.
	 */
	memset(per_cpu_state, ~0, sizeof per_cpu_state);

	perfctr_info.cpu_khz = cpu_khz;
	perfctr_info.nrcpus = smp_num_cpus;

	return 0;
}

void __exit perfctr_cpu_exit(void)
{
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		mmu_cr4_features &= ~X86_CR4_PCE;
	perfctr_cpu_exit_one(NULL);
	smp_call_function(perfctr_cpu_exit_one, NULL, 1, 1);
	perfctr_cpu_set_ihandler(NULL);
	x86_pm_exit();
	/* XXX: restart nmi watchdog? */
}

/****************************************************************
 *								*
 * Hardware reservation.					*
 *								*
 ****************************************************************/

static const char *current_service = 0;

const char *perfctr_cpu_reserve(const char *service)
{
	if( current_service )
		return current_service;
	current_service = service;
	MOD_INC_USE_COUNT;
	return 0;
}

void perfctr_cpu_release(const char *service)
{
	if( service != current_service ) {
		printk(KERN_ERR __FUNCTION__
		       ": attempt by %s to release while reserved by %s\n",
		       service, current_service);
	} else {
		perfctr_cpu_set_ihandler(NULL);
		current_service = 0;
		MOD_DEC_USE_COUNT;
	}
}
